import subprocess
import os
from scipy.signal import fftconvolve
import numpy as np
import astropy.units as U
from astropy.io import fits
from astropy import __version__ as astropy_version
from datetime import datetime
from itertools import product
from ._version import __version__ as martini_version
from warnings import warn

try:
    gc = subprocess.check_output(
        ['git', 'describe', '--always'],
        stderr=open(os.devnull, 'w'),
        cwd=os.path.dirname(os.path.realpath(__file__))
    )
except (subprocess.CalledProcessError, FileNotFoundError):
    pass
else:
    martini_version = martini_version + '_commit_' + gc.strip().decode()


class Martini():
    """
    Used for the creation of synthetic HI data cubes from simulation data.

    Usual use of martini involves first creating instances of classes from each
    of the required and optional sub-modules, then creating a Martini with
    these instances as arguments. The object can then be used to create
    synthetic observations, usually by calling 'insert_source_in_cube',
    (optionally) 'add_noise', (optionally) 'convolve_beam' and 'write_fits' in
    order.

    Parameters
    ----------
    source : an instance of a class derived from martini.source._BaseSource
        A description of the HI emitting object, including position, geometry
        and an interface to the simulation data (SPH particle masses,
        positions, etc.). Sources leveraging the simobj package for reading
        simulation data (github.com/kyleaoman/simobj) and a few test sources
        (e.g. single particle) are provided, creation of customized sources,
        for instance to leverage other interfaces to simulation data, is
        straightforward. See sub-module documentation.

    datacube : martini.DataCube instance
        A description of the datacube to create, including pixels, channels,
        sky position. See sub-module documentation.

    beam : (optional) an instance of a class derived from beams._BaseBeam
        A description of the beam for the simulated telescope. Given a
        description, either mathematical or as an image, the creation of a
        custom beam is straightforward. See sub-module documentation.

    noise : (optional) an instance of a class derived from noise._BaseNoise
        A description of the simulated noise. A simple Gaussian noise model is
        provided; implementation of other noise models is straightforward. See
        sub-module documentation.

    sph_kernel : an instance of a class derived from sph_kernels._BaseSPHKernel
        A description of the SPH smoothing kernel. The Wendland C2 kernel (used
        in EAGLE), and a point-like Dirac-delta kernel are implemented,
        implementation of other kernels is straightforward. See sub-module
        documentation.

    spectral_model : an instance of a class derived from \
    spectral_models._BaseSpectrum
        A description of the HI line produced by a particle of given
        properties. A Dirac-delta spectrum, and both fixed-width and
        temperature-dependent Gaussian line models are provided; implementing
        other models is straightforward. See sub-module documentation.

    logtag : string
        String to prepend to standard output messages.

    Returns
    -------
    out : Martini
        A Martini object configured with the specified sub-modules.

    See Also
    --------
    martini.sources
    martini.DataCube
    martini.beams
    martini.noise
    martini.sph_kernels
    martini.spectral_models

    Examples
    --------
    The following example illustrates basic use of martini, using a (very!)
    crude model of a gas disk. This example can be run by doing
    'from martini import demo; demo()'::

        from martini import Martini, DataCube
        from martini.beams import GaussianBeam
        from martini.noise import GaussianNoise
        from martini.spectral_models import GaussianSpectrum
        from martini.sph_kernels import DiracDeltaKernel
        from martini.sources import SPHSource
        import astropy.units as U
        import numpy as np

        # ------make a toy galaxy----------
        N = 1000
        phi = np.random.rand(N) * 2 * np.pi
        r = []
        for L in np.random.rand(N):
            def f(r):
                return L - .5 * (2 - np.exp(-r) * (np.power(r, 2) + 2 * r + 2))
            r.append(fsolve(f, 1.)[0])
        r = np.array(r)
        # exponential disk
        r *= 3 / np.sort(r)[N // 2]
        z = -np.log(np.random.rand(N))
        # exponential scale height
        z *= .5 / np.sort(z)[N // 2] * np.sign(np.random.rand(N) - .5)
        x = r * np.cos(phi)
        y = r * np.sin(phi)
        xyz_g = np.vstack((x, y, z)) * U.kpc
        # linear rotation curve
        vphi = 100 * r / 6.
        vx = -vphi * np.sin(phi)
        vy = vphi * np.cos(phi)
        # small pure random z velocities
        vz = (np.random.rand(N) * 2. - 1.) * 5
        vxyz_g = np.vstack((vx, vy, vz)) * U.km * U.s ** -1
        T_g = np.ones(N) * 8E3 * U.K
        mHI_g = np.ones(N) / N * 5.E9 * U.Msun
        # ~mean interparticle spacing smoothing
        hsm_g = np.ones(N) * 2 / np.sqrt(N) * U.kpc
        # ---------------------------------

        source = SPHSource(
            distance=5. * U.Mpc,
            rotation={'L_coords': (60. * U.deg, 0. * U.deg)},
            ra=0. * U.deg,
            dec=0. * U.deg,
            h=.7,
            T_g=T_g,
            mHI_g=mHI_g,
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            hsm_g=hsm_g
        )

        datacube = DataCube(
            n_px_x=128,
            n_px_y=128,
            n_channels=32,
            px_size=10. * U.arcsec,
            channel_width=10. * U.km * U.s ** -1,
            velocity_centre=source.vsys
        )

        beam = GaussianBeam(
            bmaj=30. * U.arcsec,
            bmin=30. * U.arcsec,
            bpa=0. * U.deg,
            truncate = 4.
        )

        noise = GaussianNoise(
            rms=3.E-4 * U.Jy * U.arcsec ** -2
        )

        spectral_model = GaussianSpectrum(
            sigma=7 * U.km * U.s ** -1
        )

        sph_kernel = DiracDeltaKernel()

        M = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=noise,
            spectral_model=spectral_model,
            sph_kernel=sph_kernel
        )

        M.insert_source_in_cube()
        M.add_noise()
        M.convolve_beam()
        M.write_beam_fits('testbeam.fits', channels='velocity')
        M.write_fits('testcube.fits', channels='velocity')

    The following example illustrates the usage of Martini for an IllustrisTNG
    source. The example can be run in the TNG JupyterLab environment, or in a
    local environment configured following the fiducial TNG setup
    (http://www.tng-project.org/data/docs/scripts/). It can be run by doing
    'from martini import tngdemo; tngdemo()'::

        from martini.sources import TNGSource
        from martini import DataCube, Martini
        from martini.beams import GaussianBeam
        from martini.noise import GaussianNoise
        from martini.spectral_models import GaussianSpectrum
        from martini.sph_kernels import GaussianKernel, CubicSplineKernel
        import astropy.units as U

        # This example runs in about 1 minute in the IllustrisTNG JupyterLab
        # environment.

        # Parameters myBasePath, mySnap and myId follow the usual TNG
        # conventions as in the illustris_python package.
        myBasePath = 'sims.TNG/TNG100-1/output/'
        mySnap = 99
        myId = 385350  # first central with 218 < Vmax < 220, and SFR > 1

        # The different martini sub-modules need to be initialized, see
        # https://kyleaoman.github.io/martini/build/html/martini.html for a
        # high-level overview. See the documentation for the individual sub-
        # modules for details of all configuration options. Comments below
        # highlight a few suggested best-practices specific to TNG.

        # SOURCE
        # The rotation configuration takes an inclination (here 60deg) and
        # rotation about the pole (here 0deg). The code attempts to
        # automatically align the galactic disk in the y-z plane by aligning
        # the angular momentum along the x-axis. The polar rotation is then
        # applied, and finally the disc inclined by a rotation around the
        # y-axis (the line of sight is along the x-axis). The automatic
        # alignment will work for typical reasonably isolated discs, but will
        # struggle when companions are present, when the angular momentum axis
        # is a poor tracer of the disc plane, and especially for satellites. If
        # finer control of the orientation is needed, derive the transformation
        # from the simulation box coordinates to the desired coordinates for
        # the 'observation', keeping in mind that the line of sight is along
        # the x-axis. This rotation matrix can then be passed to rotation as
        # {'rotmat': np.eye(3)} (here the identity rotation matrix used as an
        # example). A common problem in this case is deriving the inverse
        # transform instead of the forward transform, if unexpected results are
        # obtained, first try passing the transpose of the rotation matrix.
        # Note that initializing the source takes some time as the particle
        # data must be read from disk.
        source = TNGSource(
            myBasePath,
            mySnap,
            myId,
            distance=30 * U.Mpc,
            rotation={'L_coords': (60 * U.deg, 0. * U.deg)},
            ra=0. * U.deg,
            dec=0. * U.deg
        )

        # DATACUBE
        # It is usually advisable to set the centre of the cube to track the
        # centre of the source, as illustrated below. Note that the source
        # systemic velocity is set according to the distance and Hubble's law.
        # These values can instead be set explicitly, if desired. A datacube
        # with 128x128 pixels usually takes a few minutes, 1024x1024 could take
        # several hours. The number of channels has less influence on the
        # runtime. Most of the runtime is spent when M.insert_source_in_cube is
        # called below.
        datacube = DataCube(
            n_px_x=128,
            n_px_y=128,
            n_channels=32,
            px_size=10. * U.arcsec,
            channel_width=10. * U.km * U.s ** -1,
            velocity_centre=source.vsys,
            ra=source.ra,
            dec=source.dec
        )

        # BEAM
        # It is usually advisable to set the beam size to be ~3x the pixel
        # size. Note that the data cube is padded according to the size of the
        # beam, this usually results in the number of pixel rows printed in the
        # progress messages to differ from the requested dimensions. The
        # padding is required for accurate convolution with the beam, but
        # contains incorrect values after convolution and is discarded to
        # produce the final data cube of the requested size.
        beam = GaussianBeam(
            bmaj=30. * U.arcsec,
            bmin=30. * U.arcsec,
            bpa=0. * U.deg,
            truncate=3.
        )

        # NOISE
        # The noise is normally added before convolution with the beam (as
        # below in this example). The rms value passed is for the noise before
        # convolution, the rms noise in the output data cube will therefore
        # typically differ from this value.
        noise = GaussianNoise(
            rms=5.E-6 * U.Jy * U.arcsec ** -2
        )

        # SPECTRAL MODEL
        # The 'subgrid' velocity dispersion can also be fixed to a constant
        # value, e.g. sigma=7 * U.km * U.s**-1.
        spectral_model = GaussianSpectrum(
            sigma='thermal'
        )

        # SPH KERNEL
        # Since IllustrisTNG uses a moving mesh hydrodynamics solver (Arepo),
        # there are no formal SPH smoothing lengths and no specified kernel.
        # However, approximate smoothing lengths are provided by Subfind, so a
        # reasonable approximation is to use these for imaging. The lengths
        # correspond roughly to a cubic spline kernel. The implementation of
        # the cubic spline kernel included in Martini uses an approximation
        # which breaks down when the particle smoothing lengths are small
        # compared to the size of the pixels (at the distance of the source).
        # The Gaussian kernel implementation does not suffer from this
        # limitation and can be scaled to mimic the cubic spline kernel as
        # illustrated below. This solution should generally work; for
        # specialized applications, consider making a histogram of the
        # smoothing lengths of the particles in the source (in units of the
        # pixel scale) and consulting the documentation of the sph_kernels sub-
        # module to select an appropriate kernel.
        sph_kernel = GaussianKernel.mimic(
            CubicSplineKernel(rescale_sph_h=0.5),
            truncate=3
        )

        M = Martini(
            source=source,
            datacube=datacube,
            beam=beam,
            noise=noise,
            spectral_model=spectral_model,
            sph_kernel=sph_kernel
        )

        # Progress messages will be printed every printfreq rows; suppress by
        # setting to None.
        M.insert_source_in_cube(printfreq=1)
        M.add_noise()
        M.convolve_beam()

        # Two output formats are available, depending on preference. Both
        # formats are self-documenting, via FITS header keywords and HDF5
        # attributes, respectively. For HDF5 output, the beam image is included
        # in the same file.
        M.write_fits('tngdemo.fits', channels='velocity')
        M.write_beam_fits('tngdemo_beam.fits', channels='velocity')
        M.write_hdf5('tngdemo.hdf5', channels='velocity')

    """

    def __init__(self,
                 source=None,
                 datacube=None,
                 beam=None,
                 noise=None,
                 sph_kernel=None,
                 spectral_model=None,
                 logtag=''):
        self.source = source
        self.datacube = datacube
        self.beam = beam
        self.noise = noise
        self.sph_kernel = sph_kernel
        self.spectral_model = spectral_model
        self.logtag = logtag

        if self.beam is not None:
            self.beam.init_kernel(self.datacube)
            self.datacube.add_pad(self.beam.needs_pad())

        self._prune_source()

        self.spectral_model.init_spectra(self.source, self.datacube)

        return

    def convolve_beam(self):
        """
        Convolve the beam and DataCube.
        """

        if self.beam is None:
            warn('Skipping beam convolution, no beam object provided to Martini.')

        unit = self.datacube._array.unit
        for spatial_slice in self.datacube.spatial_slices():
            # use a view [...] to force in-place modification
            spatial_slice[...] = fftconvolve(
                spatial_slice, self.beam.kernel, mode='same') * unit
        self.datacube.drop_pad()
        self.datacube._array = self.datacube._array.to(
            U.Jy * U.beam**-1, equivalencies=[self.beam.arcsec_to_beam])
        return

    def add_noise(self):
        """
        Insert noise into the DataCube.
        """

        if self.noise is None:
            warn('Skipping noise, no noise object provided to Martini.')
            return

        self.datacube._array = \
            self.datacube._array + self.noise.generate(self.datacube)
        return

    def _prune_source(self):
        """
        Determines which particles cannot contribute to the DataCube and
        removes them to speed up calculation. Assumes the kernel is 0 at
        distances greater than the kernel size (which may differ from the
        SPH smoothing length).
        """

        # pixels indexed from 0 (not like in FITS!) for better use with numpy
        origin = 0
        skycoords = self.source.sky_coordinates
        particle_coords = np.vstack(
            self.datacube.wcs.sub(3).wcs_world2pix(
                skycoords.ra.to(self.datacube.units[0]),
                skycoords.dec.to(self.datacube.units[1]),
                skycoords.radial_velocity.to(self.datacube.units[2]),
                origin)) * U.pix
        # could use a function bound to source which returns the size of the
        # kernel, in case this isn't equal to the smoothing length for some
        # kernel
        sm_length = np.arctan(
            self.source.hsm_g / self.source.sky_coordinates.distance).to(
                U.pix, U.pixel_scale(self.datacube.px_size / U.pix))
        sm_range = np.ceil(sm_length * self.sph_kernel.size_in_h()).astype(int)
        spectrum_half_width = self.spectral_model.half_width(self.source) / \
            self.datacube.channel_width
        reject_conditions = (
            (particle_coords[:2] + sm_range[np.newaxis] <
             0 * U.pix).any(axis=0),
            particle_coords[0] - sm_range >
            (self.datacube.n_px_x + self.datacube.padx * 2) * U.pix,
            particle_coords[1] - sm_range >
            (self.datacube.n_px_y + self.datacube.pady * 2) * U.pix,
            particle_coords[2] + 4 * spectrum_half_width * U.pix < 0 * U.pix,
            particle_coords[2] - 4 * spectrum_half_width * U.pix >
            self.datacube.n_channels * U.pix,
        )
        reject_mask = np.zeros(particle_coords[0].shape)
        for condition in reject_conditions:
            reject_mask = np.logical_or(reject_mask, condition)
        self.source.apply_mask(np.logical_not(reject_mask))
        return

    def insert_source_in_cube(self, skip_validation=False, printfreq=100):
        """
        Populates the DataCube with flux from the particles in the source.

        Parameters
        ----------
        skip_validation : bool
            SPH kernel interpolation onto the DataCube is approximated for
            increased speed. For some combinations of pixel size, distance
            and SPH smoothing length, the approximation may break down. The
            kernel class will check whether this will occur and raise a
            RuntimeError if so. This validation can be skipped (at the cost
            of accuracy!) by setting this parameter True.

        printfreq : int or None
            Every printfreq rows a message will be printed to track progress.
            Messages completely suppressed with printfreq=None.
        """

        # pixels indexed from 0 (not like in FITS!) for better use with numpy
        origin = 0
        skycoords = self.source.sky_coordinates
        particle_coords = np.vstack(
            self.datacube.wcs.sub(3).wcs_world2pix(
                skycoords.ra.to(self.datacube.units[0]),
                skycoords.dec.to(self.datacube.units[1]),
                skycoords.radial_velocity.to(self.datacube.units[2]),
                origin)) * U.pix
        sm_length = np.arctan(
            self.source.hsm_g / self.source.sky_coordinates.distance).to(
                U.pix, U.pixel_scale(self.datacube.px_size / U.pix))
        if not skip_validation:
            self.sph_kernel.confirm_validation(sm_length)
        sm_range = np.ceil(sm_length * self.sph_kernel.size_in_h()).astype(int)

        # pixel iteration
        ij_pxs = list(
            product(
                np.arange(self.datacube._array.shape[0]),
                np.arange(self.datacube._array.shape[1])))
        if printfreq is not None:
            print(
                '  ' + self.logtag + '  [columns: {0:.0f}, rows: {1:.0f}]'
                .format(
                    self.datacube._array.shape[0],
                    self.datacube._array.shape[1]
                )
            )
        for ij_px in ij_pxs:
            ij = np.array(ij_px)[..., np.newaxis] * U.pix
            if printfreq is not None:
                if (ij[1, 0].value == 0) and (ij[0, 0].value % printfreq == 0):
                    print('  ' + self.logtag +
                          '  [row {:.0f}]'.format(ij[0, 0].value))
            mask = (np.abs(ij - particle_coords[:2]) <= sm_range).all(axis=0)
            weights = self.sph_kernel.px_weight(
                particle_coords[:2, mask] - ij,
                sm_length[mask]
            )
            self.datacube._array[ij_px[0], ij_px[1], :, 0] = \
                (self.spectral_model.spectra[mask] *
                 weights[..., np.newaxis]).sum(axis=-2)

        self.datacube._array = \
            self.datacube._array / np.power(self.datacube.px_size / U.pix, 2)
        return

    def write_fits(self, filename, channels='frequency', overwrite=True):
        """
        Output the DataCube to a FITS-format file.

        Parameters
        ----------
        filename : string
            Name of the file to write. '.fits' will be appended if not already
            present.

        channels : 'frequency' (default), or 'velocity'
            Type of units used along the spectral axis in output file.

        overwrite: bool
            Whether to allow overwriting existing files, note that the default
            is True.
        """

        self.datacube.drop_pad()
        if channels == 'frequency':
            self.datacube.freq_channels()
        elif channels == 'velocity':
            self.datacube.velocity_channels()
        else:
            raise ValueError("Martini.write_fits: Unknown 'channels' value "
                             "(use 'frequency' or 'velocity'.")

        filename = filename if filename[-5:] == '.fits' else filename + '.fits'

        wcs_header = self.datacube.wcs.to_header()
        wcs_header.rename_keyword('WCSAXES', 'NAXIS')

        header = fits.Header()
        header.append(('SIMPLE', 'T'))
        header.append(('BITPIX', 16))
        header.append(('NAXIS', wcs_header['NAXIS']))
        header.append(('NAXIS1', self.datacube.n_px_x))
        header.append(('NAXIS2', self.datacube.n_px_y))
        header.append(('NAXIS3', self.datacube.n_channels))
        header.append(('NAXIS4', 1))
        header.append(('EXTEND', 'T'))
        header.append(('CDELT1', wcs_header['CDELT1']))
        header.append(('CRPIX1', wcs_header['CRPIX1']))
        header.append(('CRVAL1', wcs_header['CRVAL1']))
        header.append(('CTYPE1', wcs_header['CTYPE1']))
        header.append(('CUNIT1', wcs_header['CUNIT1']))
        header.append(('CDELT2', wcs_header['CDELT2']))
        header.append(('CRPIX2', wcs_header['CRPIX2']))
        header.append(('CRVAL2', wcs_header['CRVAL2']))
        header.append(('CTYPE2', wcs_header['CTYPE2']))
        header.append(('CUNIT2', wcs_header['CUNIT2']))
        header.append(('CDELT3', wcs_header['CDELT3']))
        header.append(('CRPIX3', wcs_header['CRPIX3']))
        header.append(('CRVAL3', wcs_header['CRVAL3']))
        header.append(('CTYPE3', wcs_header['CTYPE3']))
        header.append(('CUNIT3', wcs_header['CUNIT3']))
        header.append(('CDELT4', wcs_header['CDELT4']))
        header.append(('CRPIX4', wcs_header['CRPIX4']))
        header.append(('CRVAL4', wcs_header['CRVAL4']))
        header.append(('CTYPE4', wcs_header['CTYPE4']))
        header.append(('CUNIT4', 'PAR'))
        header.append(('EPOCH', 2000))
        header.append(('INSTRUME', 'MARTINI', martini_version))
        # header.append(('BLANK', -32768)) #only for integer data
        header.append(('BSCALE', 1.0))
        header.append(('BZERO', 0.0))
        header.append(('DATAMAX', np.max(self.datacube._array.value)))
        header.append(('DATAMIN', np.min(self.datacube._array.value)))
        header.append(('ORIGIN', 'astropy v' + astropy_version))
        # long names break fits format, don't let the user set this
        header.append(('OBJECT', 'MOCK'))
        if self.beam is not None:
            header.append(('BPA', self.beam.bpa.to(U.deg).value))
        header.append(('OBSERVER', 'K. Oman'))
        # header.append(('NITERS', ???))
        # header.append(('RMS', ???))
        # header.append(('LWIDTH', ???))
        # header.append(('LSTEP', ???))
        header.append(('BUNIT', self.datacube._array.unit.to_string('fits')))
        # header.append(('PCDEC', ???))
        # header.append(('LSTART', ???))
        header.append(('DATE-OBS', datetime.utcnow().isoformat()[:-5]))
        # header.append(('LTYPE', ???))
        # header.append(('PCRA', ???))
        # header.append(('CELLSCAL', ???))
        if self.beam is not None:
            header.append(('BMAJ', self.beam.bmaj.to(U.deg).value))
            header.append(('BMIN', self.beam.bmin.to(U.deg).value))
        header.append(('BTYPE', 'Intensity'))
        header.append(('SPECSYS', wcs_header['SPECSYS']))

        # flip axes to write
        hdu = fits.PrimaryHDU(header=header, data=self.datacube._array.value.T)
        hdu.writeto(filename, overwrite=overwrite)

        if channels == 'frequency':
            self.datacube.velocity_channels()
        return

    def write_beam_fits(self, filename, channels='frequency', overwrite=True):
        """
        Output the beam to a FITS-format file.

        The beam is written to file, with pixel sizes, coordinate system, etc.
        similar to those used for the DataCube.

        Parameters
        ----------
        filename : string
            Name of the file to write. '.fits' will be appended if not already
            present.

        channels : 'frequency' (default), or 'velocity'
            Type of units used along the spectral axis in output file.

        overwrite: bool
            Whether to allow overwriting existing files, note that the default
            is True.

        Raises
        ------
        ValueError
            If Martini was initialized without a beam.
        """

        if self.beam is None:
            raise ValueError("Martini.write_beam_fits: Called with beam set "
                             "to 'None'.")

        if channels == 'frequency':
            self.datacube.freq_channels()
        elif channels == 'velocity':
            self.datacube.velocity_channels()
        else:
            raise ValueError("Martini.write_beam_fits: Unknown 'channels' "
                             "value (use 'frequency' or 'velocity'.")

        filename = filename if filename[-5:] == '.fits' else filename + '.fits'

        wcs_header = self.datacube.wcs.to_header()

        header = fits.Header()
        header.append(('SIMPLE', 'T'))
        header.append(('BITPIX', 16))
        # header.append(('NAXIS', self.beam.kernel.ndim))
        header.append(('NAXIS', 3))
        header.append(('NAXIS1', self.beam.kernel.shape[0]))
        header.append(('NAXIS2', self.beam.kernel.shape[1]))
        header.append(('NAXIS3', 1))
        header.append(('EXTEND', 'T'))
        header.append(('BSCALE', 1.0))
        header.append(('BZERO', 0.0))
        # this is 1/arcsec^2, is this right?
        header.append(('BUNIT', self.beam.kernel.unit.to_string('fits')))
        header.append(('CRPIX1', self.beam.kernel.shape[0] // 2 + 1))
        header.append(('CDELT1', wcs_header['CDELT1']))
        header.append(('CRVAL1', wcs_header['CRVAL1']))
        header.append(('CTYPE1', wcs_header['CTYPE1']))
        header.append(('CUNIT1', wcs_header['CUNIT1']))
        header.append(('CRPIX2', self.beam.kernel.shape[1] // 2 + 1))
        header.append(('CDELT2', wcs_header['CDELT2']))
        header.append(('CRVAL2', wcs_header['CRVAL2']))
        header.append(('CTYPE2', wcs_header['CTYPE2']))
        header.append(('CUNIT2', wcs_header['CUNIT2']))
        header.append(('CRPIX3', 1))
        header.append(('CDELT3', wcs_header['CDELT3']))
        header.append(('CRVAL3', wcs_header['CRVAL3']))
        header.append(('CTYPE3', wcs_header['CTYPE3']))
        header.append(('CUNIT3', wcs_header['CUNIT3']))
        header.append(('SPECSYS', wcs_header['SPECSYS']))
        header.append(('BMAJ', self.beam.bmaj.to(U.deg).value))
        header.append(('BMIN', self.beam.bmin.to(U.deg).value))
        header.append(('BPA', self.beam.bpa.to(U.deg).value))
        header.append(('BTYPE', 'beam    '))
        header.append(('EPOCH', 2000))
        header.append(('OBSERVER', 'K. Oman'))
        # long names break fits format
        header.append(('OBJECT', 'MOCKBEAM'))
        header.append(('INSTRUME', 'MARTINI', martini_version))
        header.append(('DATAMAX', np.max(self.beam.kernel.value)))
        header.append(('DATAMIN', np.min(self.beam.kernel.value)))
        header.append(('ORIGIN', 'astropy v' + astropy_version))

        # flip axes to write
        hdu = fits.PrimaryHDU(
            header=header, data=self.beam.kernel.value[..., np.newaxis].T)
        hdu.writeto(filename, overwrite=True)

        if channels == 'frequency':
            self.datacube.velocity_channels()
        return

    def write_hdf5(self, filename, channels='frequency', overwrite=True,
                   memmap=False, compact=False):
        """
        Output the DataCube and Beam to a HDF5-format file. Requires the h5py
        package.

        Parameters
        ----------
        filename : string
            Name of the file to write. '.hdf5' will be appended if not already
            present.

        channels : 'frequency' (default), or 'velocity'
            Type of units used along the spectral axis in output file.

        overwrite: bool
            Whether to allow overwriting existing files, note that the default
            is True.

        memmap: bool
            If True, create a file-like object in memory and return it instead
            of writing file to disk.

        compact: bool
            If True, omit pixel coordinate arrays to save disk space. In this
            case pixel coordinates can still be reconstructed from FITS-style
            keywords stored in the FluxCube attributes. Default is False.
        """

        import h5py

        self.datacube.drop_pad()
        if channels == 'frequency':
            self.datacube.freq_channels()
        elif channels == 'velocity':
            pass
        else:
            raise ValueError("Martini.write_fits: Unknown 'channels' value "
                             "(use 'frequency' or 'velocity'.")

        filename = filename if filename[-5:] == '.hdf5' else filename + '.hdf5'

        wcs_header = self.datacube.wcs.to_header()

        mode = 'w' if overwrite else 'x'
        driver = 'core' if memmap else None
        h5_kwargs = {'backing_store': False} if memmap else dict()
        f = h5py.File(filename, mode, driver=driver, **h5_kwargs)
        f['FluxCube'] = self.datacube._array.value[..., 0]
        c = f['FluxCube']
        origin = 0  # index from 0 like numpy, not from 1
        if not compact:
            xgrid, ygrid, vgrid = np.meshgrid(
                np.arange(self.datacube._array.shape[0]),
                np.arange(self.datacube._array.shape[1]),
                np.arange(self.datacube._array.shape[2])
            )
            cgrid = np.vstack((
                xgrid.flatten(),
                ygrid.flatten(),
                vgrid.flatten(),
                np.zeros(vgrid.shape).flatten()
            )).T
            wgrid = self.datacube.wcs.all_pix2world(cgrid, origin)
            ragrid = wgrid[:, 0].reshape(self.datacube._array.shape)[..., 0]
            decgrid = wgrid[:, 1].reshape(self.datacube._array.shape)[..., 0]
            chgrid = wgrid[:, 2].reshape(self.datacube._array.shape)[..., 0]
            f['RA'] = ragrid
            f['RA'].attrs['Unit'] = wcs_header['CUNIT1']
            f['Dec'] = decgrid
            f['Dec'].attrs['Unit'] = wcs_header['CUNIT2']
            f['channel_mids'] = chgrid
            f['channel_mids'].attrs['Unit'] = wcs_header['CUNIT3']
        c.attrs['AxisOrder'] = '(RA,Dec,Channels)'
        c.attrs['FluxCubeUnit'] = str(self.datacube._array.unit)
        c.attrs['deltaRA_in_RAUnit'] = wcs_header['CDELT1']
        c.attrs['RA0_in_px'] = wcs_header['CRPIX1'] - 1
        c.attrs['RA0_in_RAUnit'] = wcs_header['CRVAL1']
        c.attrs['RAUnit'] = wcs_header['CUNIT1']
        c.attrs['RAProjType'] = wcs_header['CTYPE1']
        c.attrs['deltaDec_in_DecUnit'] = wcs_header['CDELT2']
        c.attrs['Dec0_in_px'] = wcs_header['CRPIX2'] - 1
        c.attrs['Dec0_in_DecUnit'] = wcs_header['CRVAL2']
        c.attrs['DecUnit'] = wcs_header['CUNIT2']
        c.attrs['DecProjType'] = wcs_header['CTYPE2']
        c.attrs['deltaV_in_VUnit'] = wcs_header['CDELT3']
        c.attrs['V0_in_px'] = wcs_header['CRPIX3'] - 1
        c.attrs['V0_in_VUnit'] = wcs_header['CRVAL3']
        c.attrs['VUnit'] = wcs_header['CUNIT3']
        c.attrs['VProjType'] = wcs_header['CTYPE3']
        if self.beam is not None:
            c.attrs['BeamPA'] = self.beam.bpa.to(U.deg).value
            c.attrs['BeamMajor_in_deg'] = self.beam.bmaj.to(U.deg).value
            c.attrs['BeamMinor_in_deg'] = self.beam.bmin.to(U.deg).value
        c.attrs['DateCreated'] = datetime.utcnow().isoformat()[:-5]
        c.attrs['MartiniVersion'] = martini_version
        c.attrs['AstropyVersion'] = astropy_version
        if self.beam is not None:
            f['Beam'] = self.beam.kernel.value[..., np.newaxis]
            b = f['Beam']
            b.attrs['BeamUnit'] = self.beam.kernel.unit.to_string('fits')
            b.attrs['deltaRA_in_RAUnit'] = wcs_header['CDELT1']
            b.attrs['RA0_in_px'] = self.beam.kernel.shape[0] // 2
            b.attrs['RA0_in_RAUnit'] = wcs_header['CRVAL1']
            b.attrs['RAUnit'] = wcs_header['CUNIT1']
            b.attrs['RAProjType'] = wcs_header['CTYPE1']
            b.attrs['deltaDec_in_DecUnit'] = wcs_header['CDELT2']
            b.attrs['Dec0_in_px'] = self.beam.kernel.shape[1] // 2
            b.attrs['Dec0_in_DecUnit'] = wcs_header['CRVAL2']
            b.attrs['DecUnit'] = wcs_header['CUNIT2']
            b.attrs['DecProjType'] = wcs_header['CTYPE2']
            b.attrs['deltaV_in_VUnit'] = wcs_header['CDELT3']
            b.attrs['V0_in_px'] = 0
            b.attrs['V0_in_VUnit'] = wcs_header['CRVAL3']
            b.attrs['VUnit'] = wcs_header['CUNIT3']
            b.attrs['VProjType'] = wcs_header['CTYPE3']
            b.attrs['BeamPA'] = self.beam.bpa.to(U.deg).value
            b.attrs['BeamMajor_in_deg'] = self.beam.bmaj.to(U.deg).value
            b.attrs['BeamMinor_in_deg'] = self.beam.bmin.to(U.deg).value
            b.attrs['DateCreated'] = datetime.utcnow().isoformat()[:-5]
            b.attrs['MartiniVersion'] = martini_version
            b.attrs['AstropyVersion'] = astropy_version

        if channels == 'frequency':
            self.datacube.velocity_channels()
        if memmap:
            return f
        else:
            f.close()
            return

    def reset(self):
        """
        Re-initializes the DataCube with zero-values.
        """
        init_kwargs = dict(
            n_px_x=self.datacube.n_px_x,
            n_px_y=self.datacube.n_px_y,
            n_channels=self.datacube.n_channels,
            px_size=self.datacube.px_size,
            channel_width=self.datacube.channel_width,
            velocity_centre=self.datacube.velocity_centre,
            ra=self.datacube.ra,
            dec=self.datacube.dec
        )
        self.datacube.__init__(**init_kwargs)
        return



from ._martini import Martini
from ._datacube import DataCube
from ._demo import demo
from ._tngdemo import tngdemo
from ._version import __version__

__doc__ = """
Mock APERTIF-like Radio Telescope Interferometry of the Neutral ISM

MARTINI is a modular package for the creation of synthetic resolved HI line
observations (data cubes) of smoothed-particle hydrodynamics simulations of
galaxies. The various aspects of the mock-observing process are divided
logically into sub-modules handling the data cube, source, beam, noise,
spectral model and SPH kernel. MARTINI is object-oriented: each sub-module
provides a class (or classes) which can be configured as desired. For most
sub-modules, base classes are provided to allow for straightforward
customization. Instances of each sub-module class are then given as parameters
to the Martini class. A mock observation is then constructed by calling a
handful of functions to execute the desired steps in the mock-observing
process.

The package is functional and (an old version) has been used in:
Oman, K. A., Marasco, A., Navarro, J. F., Frenk, C. S., Schaye, J., Benitez-
Llambay, A. 2019, MNRAS, 482, 821.
Stable releases are available via PyPI (`pip install astromartini`) and the
numbered branches on github. The github master branch is under active
development: things will change, bugs will happen. Any feedback is greatly
appreciated.

Martini does not support use with python2.

See the help for martini.Martini for an example script to configure MARTINI
and create a datacube. This example can be run by doing:

from martini import demo
demo()

Martini has (so far) been successfully run on the output of these simulations:
 - EAGLE
 - APOSTLE
 - C-EAGLE/Hydrangea
 - Illustris
 - IllustrisTNG
 - Augiga
 - MaGICC
I attempt to support publicly available simulations with a customized source
module. If your simulation is public and not supported, please contact me at
the address below. Currently custom source modules exist for:
 - EAGLE (martini.sources.EAGLESource)
 - IllustrisTNG (martini.sources.TNGSource; also works with Illustris)

If your use of MARTINI leads to a publication, please acknowledge this and link
to the github page, ideally specifying the version used (git commit ID or
version number).

Suport available via koman@astro.rug.nl.
"""



from martini.sources import TNGSource
from martini import DataCube, Martini
from martini.beams import GaussianBeam
from martini.noise import GaussianNoise
from martini.spectral_models import GaussianSpectrum
from martini.sph_kernels import GaussianKernel, CubicSplineKernel
import astropy.units as U


def tngdemo(cubefile='tngdemo.fits', beamfile='tngdemo_beam.fits',
            hdf5file='tngdemo.hdf5'):
    # This example runs in about 1 minute in the IllustrisTNG JupyterLab
    # environment.

    # Parameters myBasePath, mySnap and myId follow the usual TNG
    # conventions as in the illustris_python package.
    myBasePath = 'sims.TNG/TNG100-1/output/'
    mySnap = 99
    myId = 385350  # first central with 218 < Vmax < 222, and SFR > 1

    # The different martini sub-modules need to be initialized, see
    # https://kyleaoman.github.io/martini/build/html/martini.html for a
    # high-level overview. See the documentation for the individual sub-
    # modules for details of all configuration options. Comments below
    # highlight a few suggested best-practices specific to TNG.

    # SOURCE
    # The rotation configuration takes an inclination (here 60deg) and
    # rotation about the pole (here 0deg). The code attempts to
    # automatically align the galactic disk in the y-z plane by aligning
    # the angular momentum along the x-axis. The polar rotation is then
    # applied, and finally the disc inclined by a rotation around the
    # y-axis (the line of sight is along the x-axis). The automatic
    # alignment will work for typical reasonably isolated discs, but will
    # struggle when companions are present, when the angular momentum axis
    # is a poor tracer of the disc plane, and especially for satellites. If
    # finer control of the orientation is needed, derive the transformation
    # from the simulation box coordinates to the desired coordinates for
    # the 'observation', keeping in mind that the line of sight is along
    # the x-axis. This rotation matrix can then be passed to rotation as
    # {'rotmat': np.eye(3)} (here the identity rotation matrix used as an
    # example). A common problem in this case is deriving the inverse
    # transform instead of the forward transform, if unexpected results are
    # obtained, first try passing the transpose of the rotation matrix.
    # Note that initializing the source takes some time as the particle
    # data must be read from disk.
    source = TNGSource(
        myBasePath,
        mySnap,
        myId,
        distance=30 * U.Mpc,
        rotation={'L_coords': (60 * U.deg, 0. * U.deg)},
        ra=0. * U.deg,
        dec=0. * U.deg
    )

    # DATACUBE
    # It is usually advisable to set the centre of the cube to track the
    # centre of the source, as illustrated below. Note that the source
    # systemic velocity is set according to the distance and Hubble's law.
    # These values can instead be set explicitly, if desired. A datacube
    # with 128x128 pixels usually takes a few minutes, 1024x1024 could take
    # several hours. The number of channels has less influence on the
    # runtime. Most of the runtime is spent when M.insert_source_in_cube is
    # called below.
    datacube = DataCube(
        n_px_x=128,
        n_px_y=128,
        n_channels=64,
        px_size=10. * U.arcsec,
        channel_width=10. * U.km * U.s ** -1,
        velocity_centre=source.vsys,
        ra=source.ra,
        dec=source.dec
    )

    # BEAM
    # It is usually advisable to set the beam size to be ~3x the pixel
    # size. Note that the data cube is padded according to the size of the
    # beam, this usually results in the number of pixel rows printed in the
    # progress messages to differ from the requested dimensions. The
    # padding is required for accurate convolution with the beam, but
    # contains incorrect values after convolution and is discarded to
    # produce the final data cube of the requested size.
    beam = GaussianBeam(
        bmaj=30. * U.arcsec,
        bmin=30. * U.arcsec,
        bpa=0. * U.deg,
        truncate=3.
    )

    # NOISE
    # The noise is normally added before convolution with the beam (as
    # below in this example). The rms value passed is for the noise before
    # convolution, the rms noise in the output data cube will therefore
    # typically differ from this value.
    noise = GaussianNoise(
        rms=5.E-6 * U.Jy * U.arcsec ** -2
    )

    # SPECTRAL MODEL
    # The 'subgrid' velocity dispersion can also be fixed to a constant
    # value, e.g. sigma=7 * U.km * U.s**-1.
    spectral_model = GaussianSpectrum(
        sigma='thermal'
    )

    # SPH KERNEL
    # Since IllustrisTNG uses a moving mesh hydrodynamics solver (Arepo),
    # there are no formal SPH smoothing lengths and no specified kernel.
    # However, approximate smoothing lengths can be estimated from the
    # Voronoi cell volumes, so a reasonable approximation is to use these
    # for imaging. The lengths correspond roughly to a cubic spline kernel.
    # The implementation of the cubic spline kernel included in Martini
    # uses an approximation which breaks down when the particle smoothing
    # lengths are small compared to the size of the pixels (at the distance
    # of the source).
    # The Gaussian kernel implementation does not suffer from this
    # limitation and can be scaled to mimic the cubic spline kernel as
    # illustrated below. This solution should generally work; for
    # specialized applications, consider making a histogram of the
    # smoothing lengths of the particles in the source (in units of the
    # pixel scale) and consulting the documentation of the sph_kernels sub-
    # module to select an appropriate kernel.
    sph_kernel = GaussianKernel.mimic(
        CubicSplineKernel(rescale_sph_h=.5),
        truncate=3
    )

    M = Martini(
        source=source,
        datacube=datacube,
        beam=beam,
        noise=noise,
        spectral_model=spectral_model,
        sph_kernel=sph_kernel
    )

    # Progress messages will be printed every printfreq rows; suppress by
    # setting to None.
    M.insert_source_in_cube(printfreq=1)
    M.add_noise()
    M.convolve_beam()

    # Two output formats are available, depending on preference. Both
    # formats are self-documenting, via FITS header keywords and HDF5
    # attributes, respectively. For HDF5 output, the beam image is included
    # in the same file.
    M.write_fits(cubefile, channels='velocity')
    M.write_beam_fits(beamfile, channels='velocity')
    try:
        M.write_hdf5(hdf5file, channels='velocity')
    except ModuleNotFoundError:
        print('h5py package not present, skipping hdf5 output demo')



import numpy as np
from os.path import join
from astropy.coordinates import CartesianRepresentation,\
    CartesianDifferential, ICRS
from astropy.coordinates.matrix_utilities import rotation_matrix
import astropy.units as U
import astropy.constants as C
from ._L_align import L_align
from ._cartesian_translation import translate, translate_d

# Extend CartesianRepresentation to allow coordinate translation
setattr(CartesianRepresentation, 'translate', translate)

# Extend CartesianDifferential to allow velocity (or other differential)
# translation
setattr(CartesianDifferential, 'translate', translate_d)


class SPHSource(object):
    """
    Class abstracting HI emission sources consisting of SPH simulation
    particles.

    This class constructs an HI emission source from arrays of SPH particle
    properties: mass, smoothing length, temperature, position, and velocity.

    Parameters
    ----------
    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
        - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element an astropy.units.Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
        - 'L_coords' : A 2-tuple containing an inclination and an azimuthal \
        angle (both astropy.units.Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    h : float
        Dimensionless hubble constant, H0 = h * 100 km / s / Mpc.

    T_g : astropy.units.Quatity, with dimensions of temperature
        Particle temperature.

    mHI_g : astropy.unit.Quantity, with dimensions of mass
        Particle HI mass.

    xyz_g : astropy.units.Quantity array of length 3, with dimensions of length
        Particle position offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    vxyz_g : astropy.units.Quantity array of length 3, with dimensions of \
    velocity
        Particle velocity offset from source centroid. Note that the 'y-z'
        plane is that eventually placed in the plane of the "sky"; 'x' is
        the axis corresponding to the "line of sight".

    hsm_g : astropy.units.Quantity, with dimensions of length
        Particle SPH smoothing lengths.

    Returns
    -------
    out : SPHSource
        An appropriately initialized SPHSource object.

    See Also
    --------
    SingleParticleSource (simplest possible implementation of a class \
    inheriting from SPHSource).
    CrossSource
    SOSource
    """

    def __init__(
            self,
            distance=3. * U.Mpc,
            vpeculiar=0. * U.km / U.s,
            rotation={'rotmat': np.eye(3)},
            ra=0.*U.deg,
            dec=0.*U.deg,
            h=.7,
            T_g=None,
            mHI_g=None,
            xyz_g=None,
            vxyz_g=None,
            hsm_g=None,
            coordinate_axis=None
    ):

        if coordinate_axis is None:
            if (xyz_g.shape[0] == 3) and (xyz_g.shape[1] != 3):
                coordinate_axis = 0
            elif (xyz_g.shape[0] != 3) and (xyz_g.shape[1] == 3):
                coordinate_axis = 1
            elif xyz_g.shape == (3, 3):
                raise RuntimeError("martini.sources.SPHSource: cannot guess "
                                   "coordinate_axis with shape (3, 3), provide"
                                   " explicitly.")
            else:
                raise RuntimeError("martini.sources.SPHSource: incorrect "
                                   "coordinate shape (not (3, N) or (N, 3)).")

        if xyz_g.shape != vxyz_g.shape:
            raise ValueError("martini.sources.SPHSource: xyz_g and vxyz_g must"
                             " have matching shapes.")
        self.h = h
        self.T_g = T_g
        self.mHI_g = mHI_g
        self.coordinates_g = CartesianRepresentation(
            xyz_g,
            xyz_axis=coordinate_axis,
            differentials={'s': CartesianDifferential(
                vxyz_g,
                xyz_axis=coordinate_axis
            )}
        )
        self.hsm_g = hsm_g

        self.npart = self.mHI_g.size

        self.ra = ra
        self.dec = dec
        self.distance = distance
        self.vpeculiar = vpeculiar
        self.rotation = rotation
        self.current_rotation = np.eye(3)
        self.rotate(**self.rotation)
        self.rotate(axis_angle=('y', self.dec))
        self.rotate(axis_angle=('z', -self.ra))
        direction_vector = np.array([
            np.cos(self.ra) * np.cos(self.dec),
            np.sin(self.ra) * np.cos(self.dec),
            np.sin(self.dec)
        ])
        distance_vector = direction_vector * self.distance
        self.translate_position(distance_vector)
        self.vsys = \
            (self.h * 100.0 * U.km * U.s ** -1 * U.Mpc ** - 1) * self.distance
        hubble_flow_vector = direction_vector * self.vsys
        vpeculiar_vector = direction_vector * self.vpeculiar
        self.translate_velocity(hubble_flow_vector + vpeculiar_vector)
        self.sky_coordinates = ICRS(self.coordinates_g)
        return

    def apply_mask(self, mask):
        """
        Remove particles from source arrays according to a mask.

        Parameters
        ----------
        mask : array-like, containing boolean-like
            Remove particles with indices corresponding to False values from
            the source arrays.
        """

        self.T_g = self.T_g[mask]
        self.mHI_g = self.mHI_g[mask]
        self.coordinates_g = self.coordinates_g[mask]
        self.sky_coordinates = ICRS(self.coordinates_g)
        self.hsm_g = self.hsm_g[mask]
        self.npart = np.sum(mask)
        if self.npart == 0:
            raise RuntimeError('No source particles in target region.')
        return

    def rotate(self, axis_angle=None, rotmat=None, L_coords=None):
        """
        Rotate the source.

        The arguments correspond to different rotation types. If supplied
        together in one function call, they are applied in order: axis_angle,
        then rotmat, then L_coords.

        Parameters
        ----------
        axis_angle : 2-tuple
            First element one of 'x', 'y', 'z' for the axis to rotate about,
            second element an astropy.units.Quantity with dimensions of angle,
            indicating the angle to rotate through.
        rotmat : (3, 3) array-like
            Rotation matrix.
        L_coords : 2-tuple
            First element containing an inclination and second element an
            azimuthal angle (both astropy.units.Quantity instances with
            dimensions of angle). The routine will first attempt to identify
            a preferred plane based on the angular momenta of the central 1/3
            of particles in the source. This plane will then be rotated to lie
            in the 'y-z' plane, followed by a rotation by the azimuthal angle
            about its angular momentum pole (rotation about 'x'), and finally
            inclined (rotation about 'y').
        """

        do_rot = np.eye(3)

        if axis_angle is not None:
            do_rot = rotation_matrix(
                axis_angle[1],
                axis=axis_angle[0]
            ).dot(do_rot)

        if rotmat is not None:
            do_rot = rotmat.dot(do_rot)

        if L_coords is not None:
            incl, az_rot = L_coords
            do_rot = L_align(self.coordinates_g.get_xyz(),
                             self.coordinates_g.differentials['s'].get_d_xyz(),
                             self.mHI_g, frac=.3, Laxis='x').dot(do_rot)
            do_rot = rotation_matrix(az_rot, axis='x').dot(do_rot)
            do_rot = rotation_matrix(incl, axis='y').dot(do_rot)

        self.current_rotation = do_rot.dot(self.current_rotation)
        self.coordinates_g = self.coordinates_g.transform(do_rot)
        return

    def translate_position(self, translation_vector):
        """
        Translate the source.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        translation_vector : astropy.units.Quantity, shape (3, ), with \
        dimensions of length
            Vector by which to offset the source particle coordinates.
        """

        self.coordinates_g = self.coordinates_g.translate(translation_vector)
        return

    def translate_velocity(self, translation_vector):
        """
        Apply an offset to the source velocity.

        Note that the "line of sight" is along the 'x' axis.

        Parameters
        ----------
        translation_vector : astropy.units.Quantity, shape (3, ), with \
        dimensions of velocity
            Vector by which to offset the source particle velocities.
        """

        self.coordinates_g.differentials['s'] = \
            self.coordinates_g.differentials['s'].translate(translation_vector)
        return

    def save_current_rotation(self, fname):
        """
        Output current rotation matrix to file.

        Parameters
        ----------
        fname : filename or file handle
            File in which to save rotation matrix.
        """

        np.savetxt(fname, self.current_rotation)
        return


class SingleParticleSource(SPHSource):
    """
    Class illustrating inheritance from martini.sources.SPHSource, creates a
    single particle test source.

    A simple test source consisting of a single particle will be created. The
    particle has a mass of 10^4 Msun, a SPH smoothing length of 1 kpc, a
    temperature of 10^4 K, a position offset by (x, y, z) = (1 pc, 1 pc, 1 pc)
    from the source centroid, a peculiar velocity of 0 km/s, and will be placed
    in the Hubble flow assuming h = 0.7 and the distance provided.

    Parameters
    ----------
    distance : astropy.units.Quantity, with units of length
        Source distance, also used to place the source in the Hubble flow
        assuming h = 0.7.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : SingleParticleSource
        An appropriately initialized SingleParticleSource object.
    """

    def __init__(
            self,
            distance=3 * U.Mpc,
            vpeculiar=0 * U.km / U.s,
            ra=0. * U.deg,
            dec=0. * U.deg
    ):

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation={'rotmat': np.eye(3)},
            ra=ra,
            dec=dec,
            h=.7,
            T_g=np.ones(1) * 1.E4 * U.K,
            mHI_g=np.ones(1) * 1.E4 * U.solMass,
            xyz_g=np.ones((1, 3)) * 1.E-3 * U.kpc,
            vxyz_g=np.zeros((1, 3)) * U.km * U.s ** -1,
            hsm_g=np.ones(1) * U.kpc
            )
        return


class CrossSource(SPHSource):
    """
    Creates a source consisting of 4 particles arrayed in an asymmetric cross.

    A simple test source consisting of four particles will be created. Each has
    a mass of 10^4 Msun, a SPH smoothing length of 1 kpc, a temperature of
    10^4 K, and will be placed in the Hubble flow assuming h=.7 and a distance
    of 3 Mpc. Particle coordinates in kpc are
    [[0,  1,  0],
    [0,  0,  2],
    [0, -3,  0],
    [0,  0, -4]]
    and velocities in km/s are
    [[0,  0,  1],
    [0, -1,  0],
    [0,  0, -1],
    [0,  1,  0]]

    Parameters
    ----------
    distance : astropy.units.Quantity, with units of length
        Source distance, also used to place the source in the Hubble flow
        assuming h = 0.7.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
        - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element an astropy.units.Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
        - 'L_coords' : A 2-tuple containing an inclination and an azimuthal \
        angle (both astropy.units.Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred \
        plane based on the angular momenta of the central 1/3 of particles \
        in the source. This plane will then be rotated to lie in the plane \
        of the "sky" ('y-z'), rotated by the azimuthal angle about its \
        angular momentum pole (rotation about 'x'), and inclined (rotation \
        about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : CrossSource
        An appropriately initialized CrossSource object.
    """

    def __init__(
            self,
            distance=3. * U.Mpc,
            vpeculiar=0 * U.km / U.s,
            rotation={'rotmat': np.eye(3)},
            ra=0. * U.deg,
            dec=0. * U.deg
    ):

        xyz_g = np.array([[0, 1, 0],
                          [0, 0, 2],
                          [0, -3, 0],
                          [0, 0, -4]]) * U.kpc,

        vxyz_g = np.array([[0, 0, 1],
                           [0, -1, 0],
                           [0, 0, -1],
                           [0, 1, 0]]) * U.km * U.s ** -1

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation={'rotmat': np.eye(3)},
            ra=ra,
            dec=dec,
            h=.7,
            T_g=np.ones(4) * 1.E4 * U.K,
            mHI_g=np.ones(4) * 1.E4 * U.solMass,
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            hsm_g=np.ones(4) * U.kpc
        )
        return


class SOSource(SPHSource):
    """
    Class abstracting HI sources using the SimObj package for interface to
    simulation data.

    This class accesses simulation data via the SimObj package
    (https://github.com/kyleaoman/simobj); see the documentation of that
    package for further configuration instructions.

    Parameters
    ----------
    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
        - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element an astropy.units.Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
        - 'L_coords' : A 2-tuple containing an inclination and an azimuthal \
        angle (both astropy.units.Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    SO_args : dict
        Dictionary of keyword arguments to pass to a call to simobj.SimObj.
        Arguments are: 'obj_id', 'snap_id', 'mask_type', 'mask_args',
        'mask_kwargs', 'configfile', 'simfiles_configfile', 'ncpu'. See simobj
        package documentation for details. Provide SO_args or SO_instance, not
        both.

    SO_instance : SimObj instance
        An initialized SimObj object. Provide SO_instance or SO_args, not both.

    Returns
    -------
    out : SOSource
        An appropriately initialized SOSource object.
    """

    def __init__(
            self,
            distance=3.*U.Mpc,
            vpeculiar=0*U.km/U.s,
            rotation={'L_coords': (60.*U.deg, 0.*U.deg)},
            ra=0.*U.deg,
            dec=0.*U.deg,
            SO_args=None,
            SO_instance=None
    ):

        from simobj import SimObj  # optional dependency for this source class

        self._SO_args = SO_args
        if (SO_args is not None) and (SO_instance is not None):
            raise ValueError('martini.source.SOSource: Provide SO_args or '
                             'SO_instance, not both.')
        if SO_args is not None:
            with SimObj(**self._SO_args) as SO:
                super().__init__(
                    distance=distance,
                    rotation=rotation,
                    ra=ra,
                    dec=dec,
                    h=SO.h,
                    T_g=SO.T_g,
                    mHI_g=SO.mHI_g,
                    xyz_g=SO.xyz_g,
                    vxyz_g=SO.vxyz_g,
                    hsm_g=SO.hsm_g
                )
        elif SO_instance is not None:
            super().__init__(
                distance=distance,
                vpeculiar=vpeculiar,
                rotation=rotation,
                ra=ra,
                dec=dec,
                h=SO_instance.h,
                T_g=SO_instance.T_g,
                mHI_g=SO_instance.mHI_g,
                xyz_g=SO_instance.xyz_g,
                vxyz_g=SO_instance.vxyz_g,
                hsm_g=SO_instance.hsm_g
            )
        else:
            raise ValueError('martini.sources.SOSource: Provide one of SO_args'
                             ' or SO_instance.')
        return


class TNGSource(SPHSource):
    """
    Class abstracting HI sources designed to run in the IllustrisTNG JupyterLab
    environment for access to simulation data. Can also be used in other
    environments, but requires that the 'illustris_python' module be
    importable, and further that the data are laid out on disk in the fiducial
    way (see: http://www.tng-project.org/data/docs/scripts/).

    Parameters
    ----------
    basePath : string
        Directory containing simulation data, for instance 'TNG100-1/output/',
        see also http://www.tng-project.org/data/docs/scripts/

    snapNum : int
        Snapshot number. In TNG, snapshot 99 is the final output. Note that
        a full snapshot (not a 'mini' snapshot, see
        http://www.tng-project.org/data/docs/specifications/#sec1a) must be
        used.

    subID : int
        Subhalo ID of the target object. Note that all particles in the FOF
        group to which the subhalo belongs are used to construct the data cube.
        This avoids strange "holes" at the locations of other subhaloes in the
        same group, and gives a more realistic treatment of foreground and
        background emission local to the source.

    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
        - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element an astropy.units.Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
        - 'L_coords' : A 2-tuple containing an inclination and an azimuthal \
        angle (both astropy.units.Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : TNGSource
        An appropriately initialized TNGSource object.
    """

    def __init__(
            self,
            basePath,
            snapNum,
            subID,
            distance=3.*U.Mpc,
            vpeculiar=0*U.km/U.s,
            rotation={'L_coords': (60.*U.deg, 0.*U.deg)},
            ra=0.*U.deg,
            dec=0.*U.deg
    ):

        # optional dependencies for this source class
        from illustris_python.groupcat import loadSingle, loadHeader
        from illustris_python.snapshot import loadSubset, getSnapOffsets
        from Hdecompose.atomic_frac import atomic_frac

        X_H = 0.76
        data_header = loadHeader(basePath, snapNum)
        data_sub = loadSingle(basePath, snapNum, subhaloID=subID)
        haloID = data_sub['SubhaloGrNr']
        fields_g = ('Masses', 'Velocities', 'InternalEnergy',
                    'ElectronAbundance', 'Density')
        subset_g = getSnapOffsets(basePath, snapNum, haloID, "Group")
        data_g = loadSubset(basePath, snapNum, 'gas', fields=fields_g,
                            subset=subset_g)
        try:
            data_g.update(loadSubset(basePath, snapNum, 'gas',
                                     fields=('CenterOfMass', ),
                                     subset=subset_g, sq=False))
        except Exception as exc:
            if ('Particle type' in exc.args[0]) and \
               ('does not have field' in exc.args[0]):
                data_g.update(loadSubset(basePath, snapNum, 'gas',
                                         fields=('Coordinates', ),
                                         subset=subset_g, sq=False))
            else:
                raise
        try:
            data_g.update(loadSubset(basePath, snapNum, 'gas',
                                     fields=('GFM_Metals', ), subset=subset_g,
                                     mdi=(0, ), sq=False))
        except Exception as exc:
            if ('Particle type' in exc.args[0]) and \
               ('does not have field' in exc.args[0]):
                X_H_g = X_H
            else:
                raise
        else:
            X_H_g = data_g['GFM_Metals']  # only loaded column 0: Hydrogen

        a = data_header['Time']
        z = data_header['Redshift']
        h = data_header['HubbleParam']
        xe_g = data_g['ElectronAbundance']
        rho_g = data_g['Density'] * 1E10 / h * U.Msun \
            * np.power(a / h * U.kpc, -3)
        u_g = data_g['InternalEnergy']  # unit conversion handled in T_g
        mu_g = 4 * C.m_p.to(U.g).value / (1 + 3 * X_H_g + 4 * X_H_g * xe_g)
        gamma = 5. / 3.  # see http://www.tng-project.org/data/docs/faq/#gen4
        T_g = (gamma - 1) * u_g / C.k_B.to(U.erg / U.K).value * 1E10 * mu_g \
            * U.K
        m_g = data_g['Masses'] * 1E10 / h * U.Msun
        # cast to float64 to avoid underflow error
        nH_g = U.Quantity(rho_g * X_H_g / mu_g, dtype=np.float64) / C.m_p
        # In TNG_corrections I set f_neutral = 1 for particles with density
        # > .1cm^-3. Might be possible to do a bit better here, but HI & H2
        # tables for TNG will be available soon anyway.
        fatomic_g = atomic_frac(
            z,
            nH_g,
            T_g,
            rho_g,
            X_H_g,
            onlyA1=True,
            TNG_corrections=True
        )
        mHI_g = m_g * X_H_g * fatomic_g
        try:
            xyz_g = data_g['CenterOfMass'] * a / h * U.kpc
        except KeyError:
            xyz_g = data_g['Coordinates'] * a / h * U.kpc
        vxyz_g = data_g['Velocities'] * np.sqrt(a) * U.km / U.s
        V_cell = data_g['Masses'] / data_g['Density'] \
            * np.power(a / h * U.kpc, 3)  # Voronoi cell volume
        r_cell = np.power(3. * V_cell / 4. / np.pi, 1. / 3.).to(U.kpc)
        hsm_g = 2.5 * r_cell  # in mind a cubic spline that =0 at h, I think
        xyz_centre = data_sub['SubhaloPos'] * a / h * U.kpc
        xyz_g -= xyz_centre
        vxyz_centre = data_sub['SubhaloVel'] * np.sqrt(a) * U.km / U.s
        vxyz_g -= vxyz_centre

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            T_g=T_g,
            mHI_g=mHI_g,
            xyz_g=xyz_g,
            vxyz_g=vxyz_g,
            hsm_g=hsm_g
        )
        return


class EAGLESource(SPHSource):
    """
    Class abstracting HI sources designed to work with publicly available
    EAGLE snapshot + group files. For file access, see
    http://icc.dur.ac.uk/Eagle/database.php.

    Parameters
    ----------
    snapPath : string
        Directory containing snapshot files.

    snapBase : string
        Filename of snapshot files, omitting portion '.X.hdf5'. Note these must
        be the subfind-processed files, usually named 
        'eagle_subfind_particles[...]'.

    groupPath : string
        Directory containing group catalog files.

    groupBase : string
        Filename of group catalog files, omitting portion '.X.hdf5'. Note
        these must be the subfind tables, usually named 
        'eagle_subfind_tab[...]'.

    fof : int
        FOF group number of the target object. Note that all particles in the
        FOF group to which the subhalo belongs are used to construct the data
        cube. This avoids strange "holes" at the locations of other subhaloes
        in the same group, and gives a more realistic treatment of foreground
        and background emission local to the source. In the EAGLE database,
        this is the 'GroupNumber'.

    sub : int
        Subfind subhalo number of the target object. For centrals the subhalo
        number is 0, for satellites >0. In the EAGLE database, this is then
        'SubGroupNumber'.

    subBoxSize : astropy.units.Quantity, with dimensions of length
        Box half-side length of a region to load around the object of interest,
        in physical (not comoving, no little h) units. This is only to avoid
        needing to load the entire particle arrays. By default set to 1 Mpc,
        which should be adequate for most galaxies.

    distance : astropy.units.Quantity, with dimensions of length
        Source distance, also used to set the velocity offset via Hubble's law.

    vpeculiar : astropy.units.Quantity, with dimensions of velocity
        Source peculiar velocity, added to the velocity from Hubble's law.

    rotation : dict
        Keys may be any combination of 'axis_angle', 'rotmat' and/or
        'L_coords'. These will be applied in this order. Note that the 'y-z'
        plane will be the one eventually placed in the plane of the "sky". The
        corresponding values:
        - 'axis_angle' : 2-tuple, first element one of 'x', 'y', 'z' for the \
        axis to rotate about, second element an astropy.units.Quantity with \
        dimensions of angle, indicating the angle to rotate through.
        - 'rotmat' : A (3, 3) numpy.array specifying a rotation.
        - 'L_coords' : A 2-tuple containing an inclination and an azimuthal \
        angle (both astropy.units.Quantity instances with dimensions of \
        angle). The routine will first attempt to identify a preferred plane \
        based on the angular momenta of the central 1/3 of particles in the \
        source. This plane will then be rotated to lie in the plane of the \
        "sky" ('y-z'), rotated by the azimuthal angle about its angular \
        momentum pole (rotation about 'x'), and inclined (rotation about 'y').

    ra : astropy.units.Quantity, with dimensions of angle
        Right ascension for the source centroid.

    dec : astropy.units.Quantity, with dimensions of angle
        Declination for the source centroid.

    Returns
    -------
    out : EAGLESource
        An appropriately initialized EAGLESource object.
    """

    def __init__(
            self,
            snapPath=None,
            snapBase=None,
            groupPath=None,
            groupBase=None,
            fof=None,
            sub=None,
            subBoxSize=1*U.Mpc,
            distance=3.*U.Mpc,
            vpeculiar=0*U.km/U.s,
            rotation={'L_coords': (60.*U.deg, 0.*U.deg)},
            ra=0.*U.deg,
            dec=0.*U.deg
    ):

        if snapPath is None:
            raise ValueError('Provide snapPath argument to EAGLESource.')
        if snapBase is None:
            raise ValueError('Provide snapBase argument to EAGLESource.')
        if groupPath is None:
            raise ValueError('Provide groupPath argument to EAGLESource.')
        if groupBase is None:
            raise ValueError('Provide groupBase argument to EAGLESource.')
        if fof is None:
            raise ValueError('Provide fof argument to EAGLESource.')
        if sub is None:
            raise ValueError('Provide sub argument to EAGLESource.')

        # optional dependencies for this source class
        from read_eagle import EagleSnapshot
        from Hdecompose.atomic_frac import atomic_frac
        import h5py

        fileN = 0
        while True:  # will break when group found or raise when out of files
            groupFile = join(groupPath, groupBase+'.{:d}.hdf5'.format(fileN))
            with h5py.File(groupFile, 'r') as g:
                mask = np.logical_and(
                    np.array(g['/Subhalo/GroupNumber']) == fof,
                    np.array(g['/Subhalo/SubGroupNumber']) == sub
                )
                if not np.sum(mask):
                    fileN += 1
                    continue  # group not in this file, try next one
                a = g['Header'].attrs['Time']
                redshift = 1 / a - 1
                h = g['Header'].attrs['HubbleParam']
                lbox = g['Header'].attrs['BoxSize'] * U.Mpc / h
                code_to_g = g['/Units'].attrs['UnitMass_in_g'] * U.g
                code_to_cm = g['/Units'].attrs['UnitLength_in_cm'] * U.cm
                code_to_cm_s = g['/Units'].attrs['UnitVelocity_in_cm_per_s'] \
                    * U.cm / U.s
                dset = g['/Subhalo/CentreOfPotential']
                aexp = dset.attrs.get('aexp-scale-exponent')
                hexp = dset.attrs.get('h-scale-exponent')
                cop = (dset[mask, :] * np.power(a, aexp) * np.power(h, hexp)
                       * code_to_cm).to(U.kpc)
                dset = g['/Subhalo/Velocity']
                aexp = dset.attrs.get('aexp-scale-exponent')
                hexp = dset.attrs.get('h-scale-exponent')
                vcent = (dset[mask, :] * np.power(a, aexp) * np.power(h, hexp)
                         * code_to_cm_s).to(U.km / U.s)
                break

        snapFile = join(snapPath, snapBase+'.0.hdf5')
        subBoxSize = (subBoxSize * h / a).to(U.Mpc).value
        centre = (cop * h / a).to(U.Mpc).value
        eagle_data = EagleSnapshot(snapFile)
        region = np.vstack((
            centre - subBoxSize,
            centre + subBoxSize
        )).T.flatten()
        eagle_data.select_region(*region)

        with h5py.File(snapFile, 'r') as f:

            fH = f['/RuntimePars'].attrs['InitAbundance_Hydrogen']
            fHe = f['/RuntimePars'].attrs['InitAbundance_Helium']
            proton_mass = f['/Constants'].attrs['PROTONMASS'] * U.g
            mu = 1 / (fH + .25 * fHe)
            gamma = f['/RuntimePars'].attrs['EOS_Jeans_GammaEffective']
            T0 = f['/RuntimePars'].attrs['EOS_Jeans_TempNorm_K'] * U.K

            def fetch(att, ptype=0):
                # gas is type 0, only need gas properties
                tmp = eagle_data.read_dataset(ptype, att)
                dset = f['/PartType{:d}/{:s}'.format(ptype, att)]
                aexp = dset.attrs.get('aexp-scale-exponent')
                hexp = dset.attrs.get('h-scale-exponent')
                return np.array(tmp, dtype='f8') * np.power(a, aexp) \
                    * np.power(h, hexp)

            ng_g = fetch('GroupNumber')
            particles = dict(
                xyz_g=(fetch('Coordinates') * code_to_cm).to(U.kpc),
                vxyz_g=(fetch('Velocity') * code_to_cm_s).to(U.km / U.s),
                T_g=fetch('Temperature') * U.K,
                hsm_g=(fetch('SmoothingLength') * code_to_cm).to(U.kpc)
            )
            rho_g = fetch('Density') * U.g * U.cm ** -3
            SFR_g = fetch('StarFormationRate')
            Habundance_g = fetch('ElementAbundance/Hydrogen')

        particles['mHI_g'] = (atomic_frac(
            redshift,
            rho_g * Habundance_g / (mu * proton_mass),
            particles['T_g'],
            rho_g,
            Habundance_g,
            onlyA1=True,
            EAGLE_corrections=True,
            SFR=SFR_g,
            mu=mu,
            gamma=gamma,
            fH=fH,
            T0=T0
        ) * code_to_g).to(U.solMass)

        mask = ng_g == fof
        for k, v in particles.items():
            particles[k] = v[mask]

        particles['xyz_g'] -= cop
        particles['xyz_g'][particles['xyz_g'] > lbox / 2.] -= lbox.to(U.kpc)
        particles['xyz_g'][particles['xyz_g'] < -lbox / 2.] += lbox.to(U.kpc)
        particles['vxyz_g'] -= vcent

        super().__init__(
            distance=distance,
            vpeculiar=vpeculiar,
            rotation=rotation,
            ra=ra,
            dec=dec,
            h=h,
            **particles
        )
        return



from setuptools import setup
import os

with open(
        os.path.join(os.path.dirname(__file__), 'martini', 'VERSION')
) as version_file:
    version = version_file.read().strip()

setup(
    name='astromartini',
    version=version,
    description='Synthetic datacube creation from simulations.',
    url='',
    author='Kyle Oman',
    author_email='koman@astro.rug.nl',
    license='GNU GPL v3',
    packages=['martini'],
    install_requires=['numpy >= 1.15.3', 'astropy >= 3.0', 'scipy'],
    include_package_data=True,
    zip_safe=False
)



from abc import ABCMeta, abstractmethod
import numpy as np
import astropy.units as U
from scipy.special import erf


class _BaseSPHKernel(object):
    """
    Abstract base class for classes implementing SPH kernels to inherit from.

    Classes inheriting from _BaseSPHKernel must implement four methods:
    'kernel', 'kernel_integral', 'validate' and 'size_in_h'.

    'kernel' should define the kernel function, normalized such that its volume
    integral is 1.

    'kernel_integral' should define the integral of the kernel over a pixel
    given the distance between the pixel centre and the particle centre, and
    the smoothing length (both in units of pixels). The integral should be
    normalized so that evaluated over the entire kernel it is equal to 1.

    'validate' should check whether any approximations converge to sufficient
    accuracy (for instance, depending on the ratio of the pixel size and
    smoothing length), and raise an error if not.

    'size_in_h' should return the maximum distance in units of the sph
    smoothing length where the kernel is non-zero.
    """

    __metaclass__ = ABCMeta

    def __init__(self, rescale_sph_h=1):
        self.rescale_sph_h = rescale_sph_h
        return

    @classmethod
    def mimic(cls, other_kernel, **kwargs):
        """
        Approximate a different kernel using this kernel.

        Note that the only class at present which has no restrictions on the
        smoothing lengths relative to the pixel size is the GaussianKernel. In
        cases where other kernels cannot be used due to such restrictions, it
        may be useful to approximate them using the GaussianKernel.

        Parameters
        ----------
        other_kernel : a class or instance inheriting from _BaseSPHKernel
            The kernel to be imitated. If the class is passed, its default
            parameters are used.

        Returns
        -------
        out : an instance of a class inheriting from _BaseSPHKernel
            An appropriately initialized object of the same type as this
            kernel.

        Examples
        --------
        The WendlandC2Kernel implementation loses accuracy when the smoothing
        lengths are small relative to the pixels. A GaussianKernel of similar
        shape can be used instead::

            kernel = GaussianKernel.mimic(WendlandC2Kernel)

        If the kernel to be mimicked must itself be rescaled for use with the
        particle SPH smoothing lengths, it can be rescaled before being
        mimicked. For instance, the CubicSplineKernel implementation here is
        non-zero in the interval [0, 2). There is an equivalent formulation of
        the same kernel which is non-zero in the interval [0, 1). To mimic this
        kernel using a GaussianKernel::

            kernel = GaussianKernel.mimic(CubicSplineKernel(rescale_sph_h=0.5))
        """

        try:
            other_kernel = other_kernel()
        except TypeError:
            pass
        if other_kernel.hscale_to_gaussian is None:
            raise ValueError(
                "{:s} cannot be rescaled, and therefore cannot be mimicked."
                .format(type(other_kernel).__name__)
            )
        if cls.hscale_to_gaussian is None:
            raise ValueError(
                "{:s} cannot be rescaled, and therefore cannot mimic other "
                "kernels.".format(cls.__name__)
            )
        return cls(
            rescale_sph_h=cls.hscale_to_gaussian
            / other_kernel.hscale_to_gaussian
            * other_kernel.rescale_sph_h,
            **kwargs
        )

    def px_weight(self, dij, h):
        """
        Calculate kernel integral using scaled smoothing lengths.

        This is the method that should be called by other modules in
        martini, rather than 'kernel_integral'.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : astropy.units.Quantity, with dimensions of pixels^-2
            Integral of smoothing kernel over pixel, per unit pixel area.
        """

        return self.kernel_integral(dij, h * self.rescale_sph_h)

    def confirm_validation(self, sm_lengths):
        """
        Verify kernel accuracy using scaled smoothing lengths.

        This is the method that should be called by other modules in
        martini, rather than 'validate'.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """

        return self.validate(sm_lengths * self.rescale_sph_h)

    def eval_kernel(self, r, h):
        """
        Evaluate the kernel, handling array casting and rescaling. Note that
        if the object was initialized with rescale_sph_h, this will be applied.

        This is the method that should be called by other modules in
        martini, rather than 'kernel'.

        Parameters
        ----------
        r : float or np.array or astropy.units.Quantity
            Distance parameter, same units as h.

        h : float or np.array or astropy.units.Quantity
            Smoothing scale parameter, same units as r.

        Returns
        -------
        out : float or np.array
            Kernel value at position(s) r / h.
        """

        q = np.array(r / h / self.rescale_sph_h)
        if isinstance(q, U.Quantity):
            q = q.to(U.dimensionless_unscaled).value
        scalar_input = q.ndim == 0
        W = self.kernel(q)
        W /= np.power(self.rescale_sph_h, 3)
        if scalar_input:
            return W.item()
        else:
            return W

    @abstractmethod
    def kernel(self, q):
        """
        Abstract method; evaluate the kernel.

        Parameters
        ----------
        q : np.array
            Dimensionless distance parameter.

        Returns
        -------
        out : np.array
            Kernel value at positions q.
        """
        pass

    @abstractmethod
    def kernel_integral(self, dij, h):
        """
        Abstract method; calculate the kernel integral over a pixel.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : astropy.units.Quantity, with dimensions of pixels^-2
            Integral of smoothing kernel over pixel, per unit pixel area.
        """

        pass

    @abstractmethod
    def validate(self, sm_lengths):
        """
        Abstract method; check conditions for validity of kernel integral
        calculation.

        Some approximations may only converge if the ratio of the pixel size
        and the smoothing length is sufficiently large, or sufficiently small.
        This method should check these conditions and raise errors when
        appropriate. The smoothing lengths are provided normalized to the pixel
        size.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """

        pass

    @abstractmethod
    def size_in_h(self):
        """
        Abstract method; return the maximum distance where the kernel is non-
        zero, in units of the SPH smoothing parameter of the particles.

        Returns
        -------
        out : float
            Maximum distance where the kernel is non-zero, in units of the SPH
            smoothing parameter.
        """
        pass


class WendlandC2Kernel(_BaseSPHKernel):
    """
    Implementation of the Wendland C2 kernel integral.

    The Wendland C2 kernel is used in the EAGLE code and derivatives (not in
    Gadget/Gadget2!). The exact integral is usually too slow to be practical;
    the implementation here approximates the kernel amplitude as constant
    across the pixel, which converges to within 1% of the exact integral
    provided the SPH smoothing lengths are at least 2 pixels in size.

    To obtain a Wendland C2 kernel which resembles a Gaussian kernel with
    sigma = 1, use rescale_sph_h = 2 * sqrt(2 * log(2)). Note that this
    scaling can be used to approximate this kernel using other kernels, or
    vice-versa.

    The WendlandC2 kernel is here defined as (q = r / h):
        W(q) = (21 / pi) * (1 - q)^4 * (4 * q + 1)
        for 0 <= q < 1
        W(q) = 0
        for q >= 1

    Parameters
    ----------
    rescale_sph_h : float
        Factor by which to rescale SPH smoothing lengths. This can be used to
        adjust particle smoothing lengths in order to approximate the kernel
        actually used in simulation with a similar kernel with different
        scaling.

    Returns
    -------
    out : WendlandC2Kernel
        An appropriately initialized WendlandC2Kernel object.
    """

    hscale_to_gaussian = 2 * np.sqrt(2 * np.log(2))

    def __init__(self, rescale_sph_h=1):
        super().__init__(rescale_sph_h=rescale_sph_h)
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The WendlandC2 kernel is here defined as (q = r / h):
        W(q) = (21 / pi) * (1 - q)^4 * (4 * q + 1)
        for 0 <= q < 1
        W(q) = 0
        for q >= 1

        Parameters
        ----------
        q : np.array
            Dimensionless distance parameter.

        Returns
        -------
        out : np.array
            Kernel value at positions q.
        """

        W = np.where(
            q < 1,
            np.power(1 - q, 4) * (4 * q + 1),
            np.zeros(q.shape)
        )
        W *= (21 / np.pi)
        return W

    def kernel_integral(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The formula used
        approximates the kernel amplitude as constant across the pixel area and
        converges to the true value within 1% for smoothing lengths >= 2
        pixels.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Approximate kernel integral over the pixel area.
        """

        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R2 = dr2 / (h * h)
        retval[R2 == 0] = 2. / 3.
        use = np.logical_and(R2 < 1, R2 != 0)
        R2 = R2[use]
        A = np.sqrt(1 - R2)
        retval[use] = 5 * R2 * R2 * (.5 * R2 + 3) * \
            np.log((1 + A) / np.sqrt(R2)) + \
            A * (-27. / 2. * R2 * R2 - 14. / 3. * R2 + 2. / 3.)
        # .2992 is normalization s.t. kernel integral = 1 for particle mass = 1
        return retval / .2992 / np.power(h, 2)

    def validate(self, sm_lengths):
        """
        Check conditions for validity of kernel integral calculation.

        Convergence within 1% of the exact integral is achieved when the
        smoothing lengths are >= 2 pixels.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.x
        """

        if (sm_lengths < 2 * U.pix).any():
            raise RuntimeError("Martini.sph_kernels.WendlandC2Kernel.validate:"
                               " SPH smoothing lengths must be >= 2 px in size"
                               " for WendlandC2 kernel integral approximation "
                               "accuracy. This check may be disabled by "
                               "calling Martini.Martini.insert_source_in_cube "
                               "with 'skip_validation=True', but use this with"
                               "care.")
        return

    def size_in_h(self):
        """
        Return the maximum distance where the kernel is non-zero.

        The WendlandC2 kernel is defined such that it reaches 0 at h=1.
        """
        return 1


class CubicSplineKernel(_BaseSPHKernel):
    """
    Implementation of the cubic spline (M4) kernel integral.

    The cubic spline is the 'classic' SPH kernel. The exact integral is usually
    too slow to be practical; the implementation here approximates the kernel
    amplitude as constant across the pixel, which converges to within 1% of
    the exact integral provided the SPH smoothing lengths are at least 2.5
    pixels in size.

    To obtain a cubic spline kernel which resembles a Gaussian kernel with
    sigma = 1, use rescale_sph_h = sqrt(2 * log(2)). Note that this scaling
    can be used to approximate this kernel using other kernels, or vice-versa.

    The cubic spline kernel is here defined as (q = r / h):
        W(q) = (2 / pi) * (1 - 1.5 * q^2 * (1 - 0.5 * q))
            for 0 <= q < 1
        W(q) = (2 / pi) * (2 - q)^3
            for 1 <= q < 2
        W(q) = 0
            for q >= 2

    Parameters
    ----------
    rescale_sph_h : float
        Factor by which to rescale SPH smoothing lengths. This can be used to
        adjust particle smoothing lengths in order to approximate the kernel
        actually used in simulation with a similar kernel with different
        scaling.

    Returns
    -------
    out : CubicSplineKernel
        An appropriately initialized CubicSplineKernel object.
    """

    hscale_to_gaussian = np.sqrt(2 * np.log(2))

    def __init__(self, rescale_sph_h=1):
        super().__init__(rescale_sph_h=rescale_sph_h)
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The cubic spline kernel is here defined as (q = r / h):
        W(q) = (2 / pi) * (1 - 1.5 * q^2 * (1 - 0.5 * q))
        for 0 <= q < 1
        W(q) = (2 / pi) * (2 - q)^3
        for 1 <= q < 2
        W(q) = 0
        for q >= 2

        Parameters
        ----------
        q : np.array
            Dimensionless distance parameter.

        Returns
        -------
        out : np.array
            Kernel value at positions q.
        """

        W = np.where(
            q < 1,
            1 - 1.5 * np.power(q, 2) + .75 * np.power(q, 3),
            .25 * np.power(2 - q, 3)
        )
        W[q > 2] = 0
        W *= 2 / np.pi
        return W

    def kernel_integral(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The formula used
        approximates the kernel amplitude as constant across the pixel area and
        converges to the true value within 1% for smoothing lengths >= 2.5
        pixels.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Approximate kernel integral over the pixel area.
        """

        dr2 = np.power(dij, 2).sum(axis=0)
        retval = np.zeros(h.shape)
        R2 = dr2 / (h * h)
        retval[R2 == 0] = 11. / 16. + .25 * .25
        case1 = np.logical_and(R2 > 0, R2 <= 1)
        case2 = np.logical_and(R2 > 1, R2 <= 4)

        R2_1 = R2[case1]
        R2_2 = R2[case2]
        A_1 = np.sqrt(1 - R2_1)
        B_1 = np.sqrt(4 - R2_1)
        B_2 = np.sqrt(4 - R2_2)
        I1 = A_1 - .5 * np.power(A_1, 3) - 1.5 * R2_1 * A_1 + 3. / 32. * A_1 \
            * (3 * R2_1 + 2) + 9. / 32. * R2_1 * R2_1 \
            * (np.log(1 + A_1) - np.log(np.sqrt(R2_1)))
        I2 = -B_2 * (3 * R2_2 + 56) / 4. \
            - 3. / 8. * R2_2 * (R2_2 + 16) \
            * np.log((2 + B_2) / np.sqrt(R2_2)) \
            + 2 * (3 * R2_2 + 4) * B_2 + 2 * np.power(B_2, 3)
        I3 = -B_1 * (3 * R2_1 + 56) / 4. + A_1 * (4 * R2_1 + 50) / 8. \
            - 3. / 8. * R2_1 * (R2_1 + 16) * np.log((2 + B_1) / (1 + A_1)) \
            + 2 * (3 * R2_1 + 4) * (B_1 - A_1) \
            + 2 * (np.power(B_1, 3) - np.power(A_1, 3))
        retval[case1] = I1 + .25 * I3
        retval[case2] = .25 * I2
        # 1.597 is normalization s.t. kernel integral = 1 for particle mass = 1
        return retval / 1.59689476201133 / np.power(h, 2)

    def validate(self, sm_lengths):
        """
        Check conditions for validity of kernel integral calculation.

        Convergence within 1% of the exact integral is achieved when the
        smoothing lengths are >= 1.0 pixels.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.x
        """

        if (sm_lengths < 1.0 * U.pix).any():
            raise RuntimeError("Martini.sph_kernels.CubicSplineKernel.validate"
                               ": SPH smoothing lengths must be >= 1.0 px in "
                               "size for cubic spline kernel integral "
                               "approximation accuracy. This check may be "
                               "disabled by calling "
                               "Martini.Martini.insert_source_in_cube with "
                               "'skip_validation=True', but use this with "
                               "care.")
        return

    def size_in_h(self):
        """
        Return the maximum distance where the kernel is non-zero.

        The cubic spline kernel is defined such that it reaches 0 at h=2.
        """
        return 1


class GaussianKernel(_BaseSPHKernel):
    """
    Implementation of a (truncated) Gaussian kernel integral.

    The 3 integrals (along dx, dy, dz) are evaluated exactly, however the
    truncation is implemented approximately in the dx and dy directions. For
    poorly sampled kernels (i.e. large pixels), the normalization is adjusted
    in order to minimize the error.

    The Gaussian kernel is here defined as (q = r / h):
    W(q) = (2 / pi^1.5) * np.exp(-q^2)
    for 0 <= q < truncate
    W(q) = 0
    for q >= truncate

    Parameters
    ----------
    truncate : float
        Number of standard deviations at which to truncate kernel (default=3).
        Truncation radii <2 may lead to large errors and are not recommended.

    rescale_sph_h : float
        Factor by which to rescale SPH smoothing lengths. This can be used to
        adjust particle smoothing lengths in order to approximate the kernel
        actually used in simulation with a similar kernel with different
        scaling.

    Returns
    -------
    out : GaussianKernel
        An appropriately initialized GaussianKernel object.
    """

    hscale_to_gaussian = 1

    def __init__(self, truncate=3, rescale_sph_h=1):
        self.truncate = truncate
        super().__init__(rescale_sph_h=rescale_sph_h)
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The Gaussian kernel is here defined as (q = r / h):
        W(q) = (2 / pi^1.5) * np.exp(-q^2)
        for 0 <= q < truncate
        W(q) = 0
        for q >= truncate

        Parameters
        ----------
        q : np.array
            Dimensionless distance parameter.

        Returns
        -------
        out : np.array
            Kernel value at positions q.
        """

        return np.where(
            q < self.truncate,
            2 * np.power(np.pi, -1.5) * np.exp(-np.power(q, 2)),
            np.zeros(q.shape)
        )

    def kernel_integral(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The 3 integrals (along dx,
        dy, dz) are evaluated exactly, however the truncation is implemented
        approximately in the dx and dy directions. For poorly sampled kernels
        (i.e. large pixels), the normalization is adjusted in order to minimize
        the error.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Kernel integral over the pixel area.
        """

        retval = .25 * (
            (erf((dij[0] + .5 * U.pix) / h) - erf((dij[0] - .5 * U.pix) / h)) *
            (erf((dij[1] + .5 * U.pix) / h) - erf((dij[1] - .5 * U.pix) / h))
        )

        # explicit truncation not required as only pixels inside
        # truncation radius should be passed, next 2 lines useful
        # for testing, however
        # dr = np.sqrt(np.power(dij, 2).sum(axis=0))
        # retval[(dr - np.sqrt(.5)) / h > self.truncate] = 0

        # empirically, removing this normalization for poorly sampled kernels
        # leads to increased accuracy
        retval[h > 2.5 * U.pix] = \
            retval[h > 2.5 * U.pix] / np.power(erf(self.truncate), 2)
        return retval * h.unit ** -2

    def validate(self, sm_lengths):

        """
        Check conditions for validity of kernel integral calculation.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """
        if self.truncate < 2:
            raise RuntimeError("Martini.sph_kernels.GaussianKernel.validate: "
                               "truncation radius of <2 may lead to errors in "
                               "mass of >1% per particle (depending on size of"
                               " smoothing kernel relative to pixels). Call "
                               "Martini.Martini.insert_source_in_cube with "
                               "'skip_validation=True' to override, at the "
                               "cost of accuracy.")
        return

    def size_in_h(self):
        """
        Return the maximum distance where the kernel is non-zero.

        The Gaussian kernel is defined such that it reaches 0 at the truncation
        radius.
        """
        return self.truncate


class DiracDeltaKernel(_BaseSPHKernel):
    """
    Implementation of a Dirac-delta kernel integral.

    The Dirac-delta kernel is here defined as (q = r / h)
    W(q) = inf
    for q == 0
    W(q) = 0
    for q != 0

    Parameters
    ----------
    rescale_sph_h : float
        Factor by which to rescale SPH smoothing lengths. This can be used to
        adjust particle smoothing lengths in order to approximate the kernel
        actually used in simulation with a similar kernel with different
        scaling.

    Returns
    -------
    out : DiracDeltaKernel
        An appropriately initialized DiracDeltaKernel object.
    """

    hscale_to_gaussian = None

    def __init__(self, rescale_sph_h=1):
        super().__init__(rescale_sph_h=rescale_sph_h)
        return

    def kernel(self, q):
        """
        Evaluate the kernel function.

        The Dirac-delta kernel is here defined as (q = r / h):
        W(q) = inf
        for q == 0
        W(q) = 0
        for q != 0

        Parameters
        ----------
        q : np.array
            Dimensionless distance parameter.

        Returns
        -------
        out : np.array
            Kernel value at positions q.
        """

        return np.where(q, np.inf * np.ones(q.shape), np.zeros(q.shape))

    def kernel_integral(self, dij, h):
        """
        Calculate the kernel integral over a pixel. The particles are
        approximated as point-like, ignoring any finite-sized kernel.
        This is a reasonable approximation provided the smoothing
        length is < 0.5 pixel in size, ideally << 1 pixel in size.

        Parameters
        ----------
        dij : astropy.units.Quantity, with dimensions of pixels
            Distances from pixel centre to particle positions, in pixels.

        h : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in pixels.

        Returns
        -------
        out : np.array
            Kernel integral over the pixel area.
        """

        return np.where((np.abs(dij) < 0.5 * U.pix).all(axis=0), 1, 0) \
            * U.pix ** -2

    def validate(self, sm_lengths):

        """
        Check conditions for validity of kernel integral calculation.

        The Dirac-delta model approaches the exact integral when the smoothing
        length is << 1 pixel in size; at a minimum the smoothing length should
        be less than half the pixel size.

        Parameters
        ----------
        sm_lengths : astropy.units.Quantity, with dimensions of pixels
            Particle smoothing lengths, in units of pixels.
        """

        if (sm_lengths > .5 * U.pix).any():
            raise RuntimeError("Martini.sph_kernels.DiracDeltaKernel.validate:"
                               " SPH smoothing lengths must be <= 1 px in size"
                               " for DiracDelta kernel to be a reasonable "
                               "approximation. Call Martini.Martini.insert_"
                               "source_in_cube with 'skip_validation=True' to "
                               "override, at the cost of accuracy.")
        return

    def size_in_h(self):
        """
        Return the maximum distance where the kernel is non-zero.

        In principle the size for a DiracDelta kernel is 0, but this would lead
        to no particles being used. Ideally we would want ~the pixel size here,
        but the sph smoothing length is acceptable.
        """
        return 1
